#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

/* #define DEBUG 1 */
/* #define DEBUG_BUSY 1 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/page.h>
#include <asm/au1000.h>

#define BUFFER_SIZE 12
#if (PAGE_SIZE < BUF_SIZE)
#define BUFFER_SIZE PAGE_SIZE   /* for not enough page size system */
#endif

#ifdef DEBUG
#define fpga_dbg(format, arg...)  printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg)
#define fpga_err(format, arg...)  printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg)
#define fpga_info(format, arg...) printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg)
#define fpga_warn(format, arg...) printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg)

#else
#define fpga_dbg(format, arg...) do {} while (0)
#define fpga_err(format, arg...)  printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg)
#define fpga_info(format, arg...) \
  if(debug > (1)) \
     printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg); \
  else \
     printk(KERN_INFO "%s : " format "\n", __FUNCTION__, ## arg)
#define fpga_warn(format, arg...) \
  if(debug > (0)) \
     printk(KERN_ERR "%s : " format "\n", __FUNCTION__, ## arg); \
  else \
     printk(KERN_WARNING "%s : " format "\n", __FUNCTION__, ## arg)
#endif

static char *in_data = NULL; /* in_data[BUFFER_SIZE] */;
static char *out_data= NULL; /* out_data[BUFFER_SIZE] */;
static unsigned long req_cmd = 0;
static unsigned long fpga_cmd = 0;
static unsigned long fpga_value = 0;

static int debug;
static int nobusy;
static int fpga_gpio_write(int num, int data);

DECLARE_WAIT_QUEUE_HEAD (fpga_wait);

static struct clientdata {
	unsigned long jiffies;
	task_queue *queue;
} fpga_data;

static struct timer_list fpga_timer;

#define AU1100_MEM_ADDR         0xB4000000
#define AU1100_GPIO2_ADDR       0xB1700000
#define AU1100_SYS_ADDR         0xB1900000
#define AU1100_CPLD_ADDR        0xB8000000
#define AU1100_EXT_ADDR         0xBA000000
#define mem(a) (AU1100_MEM_ADDR + (a))
#define sys(a) (AU1100_SYS_ADDR + (a))
#define cpld(a) (AU1100_CPLD_ADDR + (a))

#define mem_stcfg0              0x1000
#define mem_sttime0             0x1004
#define mem_staddr0             0x1008
#define mem_stcfg1              0x1010
#define mem_sttime1             0x1014
#define mem_staddr1             0x1018
#define mem_stcfg2              0x1020
#define mem_sttime2             0x1024
#define mem_staddr2             0x1028
#define mem_stcfg3              0x1030
#define mem_sttime3             0x1034
#define mem_staddr3             0x1038

#define sys_pinfunc             0x002C
#define sys_trioutrd            0x0100
#define sys_trioutclr           0x0100
#define sys_outputrd            0x0108
#define sys_outputset           0x0108
#define sys_outputclr           0x010C
#define sys_pinstaterd          0x0110
#define sys_pininputen          0x0110

#define cpld_gpio3              0x0014

/* RCE2: set 32bit mode*/
#define MEM_STCFG2_32      0x00000080 /* 80: 32bit, C0: 16bit */
#define MEM_STCFG2_16      0x000000C0 /* 80: 32bit, C0: 16bit */
#define MEM_STTIME2X     0x22080A20
#define MEM_STADDR2X     0x11A03F80

/*
 * set system bus to 16 / 32 bit mode
 */
static void fpga_bus_set(int bit)
{
	u32 v;

	if (debug) {
		v = au_readl(mem(mem_stcfg2));
		fpga_info("bit = %d, before = %08x", bit, v);
	}

	au_writel(bit == 32 ? MEM_STCFG2_32 : MEM_STCFG2_16, mem(mem_stcfg2));
	au_writel(MEM_STTIME2X, mem(mem_sttime2));
	au_writel(MEM_STADDR2X, mem(mem_staddr2));

	if (debug) {
		v = au_readl(mem(mem_stcfg2));
		fpga_info("after = %08x", v);
	}
}

/*
 * GPIO management
 *
GPIO2 => OUT
GPIO4 => OUT

GPIO8 => IN

GPIO20 => IN
GPIO21 => IN
GPIO22 => IN
GPIO23 => IN

GPIO29 => OUT
GPIO30 => OUT
GPIO31 => OUT

GPIO300 => OUT
GPIO301 => OUT
GPIO302 => OUT
GPIO303 => OUT
GPIO304 => OUT
GPIO305 => OUT
*/

static void fpga_gpio_set(void)
{
	u32 v;

	/* sys_pinfunc */
	if (debug) {
		v = au_readl(sys(sys_pinfunc));
		fpga_info("sys_pinfunc before = %08x", v);
	}

	v = au_readl(sys(sys_pinfunc));
	au_sync();
	/* set these special pin to GPIO mode
	   GPIO2:      bit 10(0)
	   GPIO8:      bit  6(0)
	   GPIO31..29: bit  5(1)
	*/
	v &= ~(0x1L << 16);
	v &= ~(0x1L << 9);
	v &= ~(0x1L << 6);
	v |=  (0x1L << 5);
	au_writel(v, sys(sys_pinfunc));
	au_sync();

	if (debug) {
		v = au_readl(sys(sys_pinfunc));
		fpga_info("sys_pinfunc after = %08x", v);
	}

	au_writel(0, sys(sys_pininputen));
	au_sync();

	/* set these GPIO in/out mode
	   GPIO2  => OUT(1)
	   GPIO4  => OUT(1)
	   GPIO8  => IN(0)
	   GPIO20 => IN(0)
	   GPIO21 => IN(0)
	   GPIO22 => IN(0)
	   GPIO23 => IN(0)
	   GPIO29 => IN(1)
	   GPIO30 => OUT(1)
	   GPIO31 => OUT(1)
	*/

	v = 0;
	v |= (0x1L << 2);
	v |= (0x1L << 4);
#ifdef DEBUG_BUSY
	v |= (0x1L << 29);
	v |= (0x1L << 8);
#endif
	v |= (0x1L << 30);
	v |= (0x1L << 31);
	au_writel(v, sys(sys_outputclr));
	au_sync();

	if (debug) {
		v = au_readl(sys(sys_outputrd));
		fpga_info("sys_outputrd before = %08x", v);
	}

#ifdef DEBUG_BUSY
	v = 0;
	v |= (0x1L << 20);
	v |= (0x1L << 21);
	v |= (0x1L << 22);
	v |= (0x1L << 23);
	au_writel(v, sys(sys_outputset));
	au_sync();

	if (debug) {
		v = au_readl(sys(sys_outputrd));
		fpga_info("sys_outputrd after = %08x", v);
	}
#endif

	/* GPIO config: sys_trioutrd, sys_trioutclr */
	if (debug) {
		v = au_readl(sys(sys_trioutrd));
		fpga_info("sys_trioutrd before = %08x", v);
	}

	v = 0;
	v |= (0x1L << 8);
	v |= (0x1L << 20);
	v |= (0x1L << 21);
	v |= (0x1L << 22);
	v |= (0x1L << 23);
	v |= (0x1L << 29);
	au_writel(v, sys(sys_trioutclr));
	au_sync();

	if (debug) {
		v = au_readl(sys(sys_trioutrd));
		fpga_info("sys_trioutrd after = %08x", v);
	}

	/* GPIO3 config: cpld_gpio3 */
	if (debug) {
		v = au_readl(cpld(cpld_gpio3));
		fpga_info("cpld_gpio3 before = %08x", v);
	}
	v = au_readl(cpld(cpld_gpio3));
	/* set these GPIO3 in/out mode
	   GPIO300 => OUT(1)
	   GPIO301 => OUT(1)
	   GPIO302 => OUT(1)
	   GPIO303 => OUT(1)
	   GPIO304 => OUT(1)
	   GPIO305 => OUT(1)
	*/
	v |= (0x1L << 16);
	v |= (0x1L << 17);
	v |= (0x1L << 18);
	v |= (0x1L << 19);
	v |= (0x1L << 20);
	v |= (0x1L << 21);
	au_writel(v, cpld(cpld_gpio3));

	if (debug) {
		v = au_readl(cpld(cpld_gpio3));
		fpga_info("sys_gpio3 after = %08x", v);
	}
	fpga_gpio_write(2, 1);
	fpga_gpio_write(4, 1);
	fpga_gpio_write(30, 1);
	fpga_gpio_write(31, 1);
}

#define SYS_GPIO1 0
#define SYS_GPIO2 200
#define CPLD_GPIO3 300

static int fpga_gpio_valid(int num)
{
	int valid;
	switch(num) {
	case 2:
	case 4:
	case 8:
	case 20:
	case 21:
	case 22:
	case 23:
	case 29:
	case 30:
	case 31:
		valid = SYS_GPIO1;
		break;
	case 300:
	case 301:
	case 302:
	case 303:
	case 304:
	case 305:
		valid = CPLD_GPIO3;
		break;
	default:
		valid = -1;
		break;
	}
	return(valid);
}

static int fpga_gpio_read(int num)
{
	int type;
	int value = -2; /* unknown error */
	u32 v;

	if ((type = fpga_gpio_valid(num)) < 0) {
		fpga_err("fpga_gpio_read, cannot read = %d", num);
		return(type);
	}
	else if(type == SYS_GPIO1) {
		v = au_readl(sys(sys_pinstaterd));
		value = (v >> num) & 0x01;
#ifdef DEBUG
		if (debug) {
			v = au_readl(sys(sys_trioutrd));
			fpga_info("sys_trioutrd = %08x", v);

			v = au_readl(sys(sys_outputrd));
			fpga_info("sys_outputrd = %08x", v);

			v = au_readl(sys(sys_pinstaterd));
			fpga_info("sys_pinstaterd = %08x", v);
		}
#endif
	}
	else if(type == CPLD_GPIO3) {
		v = au_readl(cpld(cpld_gpio3));
		value = (v >> (num - CPLD_GPIO3)) & 0x01;
	}

	if (debug) {
		fpga_info("num = %d, value = %d", num, value);
	}
	return(value);
}

static int fpga_gpio_write(int num, int data)
{
	int type = -2; /* unknown error */
	u32 v;

	data &= 0x01;
	if ((type = fpga_gpio_valid(num)) < 0) {
		fpga_err("fpga_gpio_write, cannot write = %d", num);
		return(type);
	}
	else if(type == SYS_GPIO1) {
		v = (0x01 << num);
		if (data == 1) {
			au_writel(v, sys(sys_outputset));
		}
		else {
			au_writel(v, sys(sys_outputclr));
		}
	}
	else if(type == CPLD_GPIO3) {
		v = au_readl(cpld(cpld_gpio3));
		if (data == 1) {
			v |= (0x01 << (num - CPLD_GPIO3));
		}
		else {
			v &= ~(0x01 << (num - CPLD_GPIO3));
		}
		au_writel(v, cpld(cpld_gpio3));
	}
	return(type);
}

/*
 * RD IO
 */
static u32 fpga_rd_read(void)
{
	return(au_readl(AU1100_EXT_ADDR));
}

static void fpga_rd_write(u32 data)
{
	au_writel(data, AU1100_EXT_ADDR);
}

/*
 * timer management
 */
static void fpga_timedout(unsigned long ptr)
{
	fpga_dbg("** timedout **");
        wake_up_interruptible(&fpga_wait);  /* awake the process */
}

/* sleep while ms millisecond */
static unsigned int fpga_sleep(unsigned long ms)
{
	fpga_dbg("** strat **");
        fpga_data.jiffies = jiffies;
        fpga_data.queue = NULL;      /* don't requeue */

        init_timer(&fpga_timer);     /* init the timer structure */
        fpga_timer.function = fpga_timedout;

        fpga_dbg("**init_timer & timer_function done = %ld ms", ms);

        fpga_timer.data = (unsigned long)&fpga_data;
        /* fpga_timer.expires = jiffies + HZ; * one second */
        fpga_timer.expires = jiffies + (ms / 10); /* ms millisecond */

        add_timer(&fpga_timer);
        interruptible_sleep_on(&fpga_wait);
        del_timer_sync(&fpga_timer);  /* in case a signal woke us up */

	fpga_dbg("** end **");
        return jiffies;
}

/*
 * fpga_read_proc -- called when user reading
 */
static
int fpga_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
    int return_length = 0;
    int req;

    fpga_dbg("count = %d, off = %d", count, (int) offset);

    if (offset == 0) {
#ifdef DEBUG
	    while(0 && (req = fpga_gpio_read(8))) { /* REQ ? */
#else
	    while((req = fpga_gpio_read(8))) { /* REQ ? */
#endif
		    fpga_sleep(10); /* sleep 10ms */
	    }
	    req_cmd  = fpga_gpio_read(20);
	    req_cmd |= fpga_gpio_read(21) << 1;
	    req_cmd |= fpga_gpio_read(22) << 2;
	    req_cmd |= fpga_gpio_read(23) << 3;
#ifdef DEBUG
	    /* for test */
	    req_cmd = fpga_cmd;
	    if (debug > 3 && req_cmd > 0 && req_cmd < 20) { /* debug > 3 => sleep debug */
		    (void) fpga_sleep(req_cmd * 1000); /* This is a sleep test. */
	    }
	    if (req_cmd & 0x01) {
		    fpga_gpio_write(2, 1);
		    fpga_dbg("fpga_gpio_write(2, 1);");
	    }
	    else {
		    fpga_gpio_write(2, 0);
		    fpga_dbg("fpga_gpio_write(2, 0);");
	    }
	    if (req_cmd & 0x02) {
		    fpga_gpio_write(4, 1);
		    fpga_dbg("fpga_gpio_write(4, 1);");
	    }
	    else {
		    fpga_gpio_write(4, 0);
		    fpga_dbg("fpga_gpio_write(4, 0);");
	    }
	    if (req_cmd & 0x04) {
		    fpga_gpio_write(30, 1);
		    fpga_dbg("fpga_gpio_write(30, 1);");
	    }
	    else {
		    fpga_gpio_write(30, 0);
		    fpga_dbg("fpga_gpio_write(30, 0);");
	    }
	    if (req_cmd & 0x08) {
		    fpga_gpio_write(31, 1);
		    fpga_dbg("fpga_gpio_write(31, 1);");
	    }
	    else {
		    fpga_gpio_write(31, 0);
		    fpga_dbg("fpga_gpio_write(31, 0);");
	    }
	    if (req_cmd & 0x10) {
		    fpga_gpio_write(305, 1);
		    fpga_dbg("fpga_gpio_write(305, 1);");
	    }
	    else {
		    fpga_gpio_write(305, 0);
		    fpga_dbg("fpga_gpio_write(305, 0);");
	    }
#endif
      memset(in_data, 0, BUFFER_SIZE);
      return_length = sprintf(in_data, "%lu\n", req_cmd);
      memcpy(buf, in_data, return_length);

      fpga_dbg("cmd = %lu", simple_strtoul(in_data, NULL, 10));
    }
    *eof = 1;
    *start = buf + offset;
    return return_length;
}

/*
 * fpga_dt_read_proc -- called when user reading
 */
static
int fpga_dt_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	int return_length = 0;
	u32 v;

	fpga_dbg("count = %d, off = %d", count, (int) offset);

	if (offset == 0) {
		v = fpga_rd_read();
		memset(in_data, 0, BUFFER_SIZE);
		return_length = sprintf(in_data, "%08x\n", v);
		memcpy(buf, in_data, return_length);

		fpga_dbg("data = %08x", v);
	}
	*eof = 1;
	*start = buf + offset;
	return return_length;
}

/*
 * fpga_bit_read_proc -- called when user reading
 */
static
int fpga_bit_read_proc(int gpio, char *buf, char **start, off_t offset,
                   int count, int *eof)
{
	int return_length = 0;
	int v;

	fpga_dbg("gpio = %d, count = %d, off = %d", gpio, count, (int) offset);

	if (offset == 0) {
		switch(gpio) {
			/* those for write debug */
		case 2:
		case 4:
		case 31:
		case 30:
			/* those for read GPIO */
		case 8:
		case 20:
		case 21:
		case 22:
		case 23:
		case 29:
			v = fpga_gpio_read(gpio);
			break;
		default:
			v = 0;
			break;
		}
		memset(in_data, 0, BUFFER_SIZE);
		return_length = sprintf(in_data, "%d\n", v);
		memcpy(buf, in_data, return_length);

		fpga_dbg("cmd = %lu", simple_strtoul(in_data, NULL, 10));
	}
	*eof = 1;
	*start = buf + offset;
	return return_length;
}

static
int fpga_sda_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	return(fpga_bit_read_proc(2, buf, start, offset, count, eof));
}

static
int fpga_exa_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	return(fpga_bit_read_proc(4, buf, start, offset, count, eof));
}

static
int fpga_mrw_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	return(fpga_bit_read_proc(31, buf, start, offset, count, eof));
}

static
int fpga_exrw_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	return(fpga_bit_read_proc(30, buf, start, offset, count, eof));
}

static
int fpga_busy_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	return(fpga_bit_read_proc(29, buf, start, offset, count, eof));
}

static
int fpga_req_read_proc(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	return(fpga_bit_read_proc(8, buf, start, offset, count, eof));
}

/*
 * fpga_write_proc -- called when user writing
 */
int fpga_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	struct inode * inode = file->f_dentry->d_inode;
	struct proc_dir_entry * dp;
	int length;
	unsigned long v;
 
	dp = (struct proc_dir_entry *) inode->u.generic_ip;
	fpga_dbg("%s: count = %lu", dp->name, count);

	length = count > BUFFER_SIZE ? BUFFER_SIZE : count; /* limit of max size */
	copy_from_user(out_data, buf, length);

	fpga_cmd = simple_strtoul(out_data, NULL, 10);
#ifdef OLD_BUSY
	while(!nobusy && fpga_gpio_read(29) == 0) { /* is busy ? */
		if (debug > 1) {
			fpga_info("gpio_read(29) is busy, call schedule()...");
		}
		schedule();
	}
#endif
	v = (fpga_cmd >> 4) & 0x01;
	fpga_gpio_write(304, v);
	v = (fpga_cmd >> 3) & 0x01;
	fpga_gpio_write(303, v);
	v = (fpga_cmd >> 2) & 0x01;
	fpga_gpio_write(302, v);
	v = (fpga_cmd >> 1) & 0x01;
	fpga_gpio_write(301, v);
	v = (fpga_cmd >> 0) & 0x01;
	fpga_gpio_write(300, v);

	/* command request to go */
	fpga_gpio_write(305, 0);
	fpga_gpio_write(305, 1);

	fpga_dbg("cmd = %ld", fpga_cmd);

	return length;
}

/*
 * fpga_dt_write_proc -- called when user writing
 */
int fpga_dt_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	struct inode * inode = file->f_dentry->d_inode;
	struct proc_dir_entry * dp;
	int length;
 
	dp = (struct proc_dir_entry *) inode->u.generic_ip;
	fpga_dbg("%s, count = %lu", dp->name, count);

	length = count > BUFFER_SIZE ? BUFFER_SIZE : count; /* limit of max size */	   
	copy_from_user(out_data, buf, length);

	fpga_value = simple_strtoul(out_data, NULL, 16);
	fpga_rd_write(fpga_value);
	fpga_dbg("data = %08lx", fpga_value);

	return length;
}

/*
 * fpga_bit_write_proc -- called when user writing
 */
int fpga_bit_write_proc(int gpio, struct file *file, const char *buf,
                     unsigned long count)
{
	struct inode * inode = file->f_dentry->d_inode;
	struct proc_dir_entry * dp;
	int length, v, error;

	dp = (struct proc_dir_entry *) inode->u.generic_ip; 
	fpga_dbg("%s: count = %lu", dp->name, count);

	length = count > 2 ? 2 : count; /* limit of max size */
	copy_from_user(out_data, buf, length);

	v = simple_strtoul(out_data, NULL, 10);

	switch(gpio) {
	case 2:
	case 4:
	case 31:
	case 30:
		break;
#ifdef DEBUG_BUSY
	case 29:
	case 8:
		break;
#endif
	default:
		fpga_err("invalid file name = (%s) %d, %d",
			 dp->name, gpio, v);
		return length;
	}

	/* NEW_BUSY_WAITING ... */
	while(!nobusy && fpga_gpio_read(29) == 0) { /* is busy ? */
		if (debug > 1) {
			fpga_info("gpio_read(29) is busy, call schedule()...");
		}
		schedule();
	}
	/* end of NEW_BUSY_WAITING */

	if ((error = fpga_gpio_write(gpio, v)) < 0) {
		fpga_err("fpga_gpio_write error(%d) = %d, %d",
			 error, gpio, v);
	}
	fpga_dbg("data = %d", v);

	return length;
}

/*
 * fpga_sda_write_proc -- called when user writing
 */
int fpga_sda_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	return(fpga_bit_write_proc(2, file, buf, count));
}

/*
 * fpga_exa_write_proc -- called when user writing
 */
int fpga_exa_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	return(fpga_bit_write_proc(4, file, buf, count));
}

/*
 * fpga_mrw_write_proc -- called when user writing
 */
int fpga_mrw_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	return(fpga_bit_write_proc(31, file, buf, count));
}

/*
 * fpga_exrw_write_proc -- called when user writing
 */
int fpga_exrw_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	return(fpga_bit_write_proc(30, file, buf, count));
}

/*
 * fpga_busy_write_proc -- called when user writing
 */
int fpga_busy_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	return(fpga_bit_write_proc(29, file, buf, count));
}

/*
 * fpga_req_write_proc -- called when user writing
 */
int fpga_req_write_proc(struct file *file, const char *buf,
                     unsigned long count, void *data)
{
	return(fpga_bit_write_proc(8, file, buf, count));
}

/*
 * creating and removing /proc/fpga* files
 */
static void fpga_create_proc(void)
{
	struct proc_dir_entry *entry;
	entry = create_proc_entry("fpga", 0, 0); /* "/proc/fpga" registration */
	entry->read_proc = fpga_read_proc;       /* read routine */
	entry->write_proc = fpga_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_dt", 0, 0); /* "/proc/fpga_dt" registration */
	entry->read_proc = fpga_dt_read_proc;       /* read routine */
	entry->write_proc = fpga_dt_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_sda", 0, 0); /* "/proc/fpga_sda" registration */
	entry->read_proc = fpga_sda_read_proc;       /* read routine */
	entry->write_proc = fpga_sda_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_exa", 0, 0); /* "/proc/fpga_exa" registration */
	entry->read_proc = fpga_exa_read_proc;       /* read routine */
	entry->write_proc = fpga_exa_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_mrw", 0, 0); /* "/proc/fpga_mrw" registration */
	entry->read_proc = fpga_mrw_read_proc;       /* read routine */
	entry->write_proc = fpga_mrw_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_exrw", 0, 0); /* "/proc/fpga_exrw" registration */
	entry->read_proc = fpga_exrw_read_proc;       /* read routine */
	entry->write_proc = fpga_exrw_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_busy", 0, 0); /* "/proc/fpga_busy" registration */
	entry->read_proc = fpga_busy_read_proc;       /* read routine */
	entry->write_proc = fpga_busy_write_proc;     /* write routine */

	entry = create_proc_entry("fpga_req", 0, 0); /* "/proc/fpga_req" registration */
	entry->read_proc = fpga_req_read_proc;       /* read routine */
	entry->write_proc = fpga_req_write_proc;     /* write routine */
}

static void fpga_remove_proc(void)
{
	remove_proc_entry("fpga_req", NULL);
	remove_proc_entry("fpga_busy", NULL);
	remove_proc_entry("fpga_exrw", NULL);
	remove_proc_entry("fpga_mrw", NULL);
	remove_proc_entry("fpga_exa", NULL);
	remove_proc_entry("fpga_sda", NULL);
	remove_proc_entry("fpga_dt", NULL);
	remove_proc_entry("fpga", NULL);
}

/*
 * init and exit
 */
static
int __init fpga_init_module(void)
{
	in_data = (char *) vmalloc(BUFFER_SIZE);
	if (in_data == NULL) {
		fpga_err("FPGA in_data, cannot vmalloc = %d", BUFFER_SIZE);
		return 0;
	}
	out_data = (char *) vmalloc(BUFFER_SIZE);
	if (out_data == NULL) {
		fpga_err("FPGA out_data, cannot vmalloc = %d", BUFFER_SIZE);
		return 0;
	}
	fpga_bus_set(32);
	fpga_gpio_set();
	fpga_create_proc();
	printk("<1>FPGA, init\n");
	return 0;
}

static
void __exit fpga_cleanup_module(void)
{
	printk("<1>FPGA cleanup\n");
	if (in_data != NULL) {
		vfree(in_data);
	}
	if (out_data != NULL) {
		vfree(out_data);
	}
	fpga_remove_proc();
	fpga_bus_set(16);
}

MODULE_DESCRIPTION("FPGA Module");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Device Drivers Ltd.");
MODULE_PARM(debug,       "i");
MODULE_PARM(nobusy,       "i");
MODULE_PARM_DESC(debug, "FPGA debug level (0-6)");
MODULE_PARM_DESC(nobusy, "FPGA ignore BUSY state");

module_init(fpga_init_module);
module_exit(fpga_cleanup_module);
